"""Tests for advanced author features: affiliation numbering and note symbols."""

from conftest import compare_have_want


def test_two_authors_same_affiliation():
    """Two authors with same affiliation: no superscripts (only 1 unique)."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :affiliation: MIT
        }
        ::

        :author:{
          :name: Bob
          :affiliation: MIT
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice</p>

        <p>MIT</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob</p>

        <p>MIT</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_authors_different_affiliations():
    """Two authors with different affiliations should get different numbers."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :affiliation: MIT
        }
        ::

        :author: {
          :name: Bob
          :affiliation: Harvard
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice<sup>1</sup></p>

        <p><sup>1</sup>MIT</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob<sup>2</sup></p>

        <p><sup>2</sup>Harvard</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_three_authors_mixed_affiliations():
    """Three authors: two share affiliation, one different."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :affiliation: MIT
        }
        ::

        :author:{
          :name: Bob
          :affiliation: Harvard
        }
        ::

        :author:{
          :name: Carol
          :affiliation: MIT
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice<sup>1</sup></p>

        <p><sup>1</sup>MIT</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob<sup>2</sup></p>

        <p><sup>2</sup>Harvard</p>
        </div>
        </div>

        <div class="author" data-nodeid="3">

        <div class="paragraph">

        <p>Carol<sup>1</sup></p>

        <p><sup>1</sup>MIT</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_author_with_no_affiliation():
    """Author without affiliation should not get a number."""
    compare_have_want(
        have="""\
        # Test

        :author:{
          :name: Alice
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_authors_same_note():
    """Two authors with same note: no superscripts (only 1 unique)."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :author-note: Equal contribution
        }
        ::

        :author: {
          :name: Bob
          :author-note: Equal contribution
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice</p>

        <p>Equal contribution</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob</p>

        <p>Equal contribution</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_authors_different_notes():
    """Two authors with different notes should get different symbols."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :author-note: Equal contribution
        }
        ::

        :author: {
          :name: Bob
          :author-note: Now at MIT
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice<sup>*</sup></p>

        <p><sup>*</sup>Equal contribution</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob<sup>†</sup></p>

        <p><sup>†</sup>Now at MIT</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_author_with_affiliation_and_note():
    """Single author: no superscripts (only 1 unique affiliation and note)."""
    compare_have_want(
        have="""\
        # Test

        :author: {
          :name: Alice
          :affiliation: MIT
          :author-note: Equal contribution
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice</p>

        <p>MIT</p>

        <p>Equal contribution</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_many_notes_symbol_progression():
    """Test symbol progression: *, †, ‡, §, ¶, ‖, **, †† (with >5 authors collapsed)."""
    compare_have_want(
        have="""\
        # Test

        :author:{
          :name: A1
          :author-note: Note 1
        }
        ::

        :author:{
          :name: A2
          :author-note: Note 2
        }
        ::

        :author:{
          :name: A3
          :author-note: Note 3
        }
        ::

        :author:{
          :name: A4
          :author-note: Note 4
        }
        ::

        :author:{
          :name: A5
          :author-note: Note 5
        }
        ::

        :author:{
          :name: A6
          :author-note: Note 6
        }
        ::

        :author:{
          :name: A7
          :author-note: Note 7
        }
        ::

        :author:{
          :name: A8
          :author-note: Note 8
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="authors-container">

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>A1<sup>*</sup></p>

        <p><sup>*</sup>Note 1</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>A2<sup>†</sup></p>

        <p><sup>†</sup>Note 2</p>
        </div>
        </div>

        <div class="author" data-nodeid="3">

        <div class="paragraph">

        <p>A3<sup>‡</sup></p>

        <p><sup>‡</sup>Note 3</p>
        </div>
        </div>

        <div class="author author-hidden author-toggleable" data-nodeid="4">

        <div class="paragraph">

        <p>A4<sup>§</sup></p>

        <p><sup>§</sup>Note 4</p>
        </div>
        </div>

        <div class="author author-hidden author-toggleable" data-nodeid="5">

        <div class="paragraph">

        <p>A5<sup>¶</sup></p>

        <p><sup>¶</sup>Note 5</p>
        </div>
        </div>

        <div class="author author-hidden author-toggleable" data-nodeid="6">

        <div class="paragraph">

        <p>A6<sup>‖</sup></p>

        <p><sup>‖</sup>Note 6</p>
        </div>
        </div>

        <button class="toggle-authors">Show/hide full author information</button>

        <div class="author" data-nodeid="7">

        <div class="paragraph">

        <p>A7<sup>**</sup></p>

        <p><sup>**</sup>Note 7</p>
        </div>
        </div>

        <div class="author" data-nodeid="8">

        <div class="paragraph">

        <p>A8<sup>††</sup></p>

        <p><sup>††</sup>Note 8</p>
        </div>
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_complex_author_combination():
    """Complex: 2 affiliations (show labels), 1 note (no labels)."""
    compare_have_want(
        have="""\
        # Test

        :author:{
          :name: Alice
          :affiliation: MIT
          :orcid: 0000-0001-1111-1111
          :author-note: Equal contribution
        }
        ::

        :author:{
          :name: Bob
          :affiliation: Harvard
          :author-note: Equal contribution
        }
        ::

        :author:{
          :name: Carol
          :affiliation: MIT
          :orcid: 0000-0002-2222-2222
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Test</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Alice<sup>1</sup></p>

        <p><sup>1</sup>MIT</p>

        <p>0000-0001-1111-1111</p>

        <p>Equal contribution</p>
        </div>
        </div>

        <div class="author" data-nodeid="2">

        <div class="paragraph">

        <p>Bob<sup>2</sup></p>

        <p><sup>2</sup>Harvard</p>

        <p>Equal contribution</p>
        </div>
        </div>

        <div class="author" data-nodeid="3">

        <div class="paragraph">

        <p>Carol<sup>1</sup></p>

        <p><sup>1</sup>MIT</p>

        <p>0000-0002-2222-2222</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""
util.py
-------

Utilities.

"""

import sys
import textwrap
from collections.abc import Generator
from typing import Any

from pygments import highlight
from pygments.formatters import HtmlFormatter  # pylint: disable=no-name-in-module
from pygments.lexers import get_lexer_by_name


# pylint: disable-next=too-few-public-methods
class RSMPygmentsFormatter(HtmlFormatter):
    def wrap(
        self, source: Generator[tuple[int, str], None, None]
    ) -> Generator[tuple[int, str], None, None]:
        return self._wrap_code(source)

    def _wrap_code(
        self, source: Generator[tuple[int, str], None, None]
    ) -> Generator[tuple[int, str], None, None]:
        yield from source

    def _wrap_div(
        self, source: Generator[tuple[int, str], None, None]
    ) -> Generator[tuple[int, str], None, None]:
        yield from source


def highlight_code(source: str, lang: str) -> str:
    # Pygments PythonLexer hits recursion limits on Python 3.10-3.12
    # Temporarily increase limit for lexer initialization
    old_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(3000)
    try:
        highlighted = highlight(
            source,
            get_lexer_by_name(lang),
            RSMPygmentsFormatter(),
        )
        return highlighted  # type: ignore[no-any-return]
    finally:
        sys.setrecursionlimit(old_limit)


class EscapedString:
    def __init__(self, src: str = "", chars: str | None = None):
        self.escape_chars = set() if chars is None else set(chars)
        self._src = str(src)

    def __contains__(self, sub: str) -> bool:
        return sub in self._src

    def __len__(self) -> int:
        return len(self._src)

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return f"{cls}({self.escape_chars}, {textwrap.shorten(self._src, 60)})"

    def __str__(self) -> str:
        return self._src

    def escape(self) -> str:
        ret = self._src
        for char in self.escape_chars:
            ret = ret.replace(f"\\{char}", char)
        return ret

    def __getitem__(self, _slice: int | slice) -> "EscapedString":
        return self.__class__(self._src[_slice])

    def __add__(self, other: str) -> str:
        return self._src + other

    def __radd__(self, other: str) -> str:
        return other + self._src

    def __eq__(self, other: Any) -> bool:
        return self._src == other  # type: ignore[no-any-return]

    def __hash__(self) -> int:
        return hash(self._src)

    def format(self, *args: Any, **kwargs: Any) -> str:
        return self._src.format(*args, **kwargs)

    def strip(self, chars: str | None = None, /) -> "EscapedString":
        return self.__class__(self._src.strip(chars))

    def lstrip(self, chars: str | None = None, /) -> "EscapedString":
        return self.__class__(self._src.lstrip(chars))

    def rstrip(self, chars: str | None = None, /) -> "EscapedString":
        return self.__class__(self._src.rstrip(chars))

    def find(
        self,
        sub: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> int:
        if end is None:
            end = len(self._src)
        if not skip_escaped or sub not in self.escape_chars:
            return self._src.find(sub, start, end)
        index = self._src.find(sub, start, end)
        if index < 1:
            # If sub was found at index 0, it cannot have been escaped since there is
            # nothing behind it!  If the index was negative, it was not found.  Either
            # way, we are done.
            return index
        while self._src[index - 1] == "\\":
            index = self._src.find(sub, index + 1, end)
            if index < 0:
                return index
        return index

    def rfind(
        self,
        sub: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> int:
        if end is None:
            end = len(self._src)
        if not skip_escaped or sub not in self.escape_chars:
            return self._src.rfind(sub, start, end)
        raise Exception("Leo should have implemented this but didn't")

    def index(
        self,
        sub: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> int:
        if end is None:
            end = len(self._src)
        if not skip_escaped or sub not in self.escape_chars:
            return self._src.index(sub, start, end)
        index = self._src.index(sub, start, end)
        if index == 0:
            # If sub was found at index 0, it cannot have been escaped since there is
            # nothing behind it!  Unlike find(), index() never returns -1 (it raises an
            # exception instead), so no need to check for that.
            return index
        while self._src[index - 1] == "\\":
            index = self._src.find(sub, index + 1, end)
        return index

    def rindex(
        self,
        sub: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> int:
        if end is None:
            end = len(self._src)
        if not skip_escaped or sub not in self.escape_chars:
            return self._src.rindex(sub, start, end)
        raise Exception("Leo should have implemented this but didn't")

    def replace(
        self, old: str, new: str, count: int = -1, /, skip_escaped: bool = True
    ) -> "EscapedString":
        if not skip_escaped or old not in self.escape_chars:
            return self.__class__(self._src.replace(old, new, count))
        raise Exception("Leo should have implemented this but didn't")

    def startswith(
        self,
        prefix: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> bool:
        if end == -1:
            end = len(self._src)
        if len(prefix) == 1:
            return self._src.startswith(prefix, start, end)
        if not skip_escaped or prefix not in self.escape_chars:
            return self._src.startswith(prefix, start, end)
        raise Exception("Leo should have implemented this but didn't")

    def endswith(
        self,
        prefix: str,
        start: int = 0,
        end: int | None = None,
        skip_escaped: bool = True,
    ) -> bool:
        if end is None:
            end = len(self._src)
        if not skip_escaped or prefix not in self.escape_chars:
            return self._src.endswith(prefix, start, end)
        raise Exception("Leo should have implemented this but didn't")

    def split(
        self,
        /,
        sep: str | None = None,
        maxsplit: int = -1,
        skip_escaped: bool = True,
    ) -> list[str]:
        if not skip_escaped or sep not in self.escape_chars:
            return self._src.split(sep, maxsplit)
        runs = []
        prev = 0
        curr = 0
        while curr < len(self._src):
            if self._src[curr] in self.escape_chars and (
                curr > 0 and self._src[curr - 1] != "\\"
            ):
                runs.append(self._src[prev + 1 : curr - 1])
                prev = curr
            curr += 1
        runs.append(self._src[prev + 1 : curr - 1])

        return runs



"""Tests for standalone HTML output mode.

When standalone=True, the RSM builder produces a single HTML file that can be
opened directly in a browser without requiring a web server. This is achieved by:
1. Using CDN URLs for third-party libraries (jQuery, Tooltipster)
2. Using absolute URLs to the Studio backend for RSM-specific assets
"""

import rsm


class TestStandaloneMode:
    """Test standalone=True produces self-contained HTML."""

    def test_make_accepts_standalone_parameter(self):
        """Test that rsm.build() accepts standalone parameter."""
        source = ":rsm:\n\nHello world.\n\n::"
        # Should not raise
        result = rsm.build(source, handrails=False, lint=False, standalone=True)
        assert "<html" in result.lower()

    def test_standalone_false_uses_absolute_paths(self):
        """Test that standalone=False (default) uses /static/ paths."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=False)

        assert 'href="/static/rsm.css"' in result
        assert 'href="/static/tooltipster.bundle.css"' in result
        assert 'src="/static/jquery-3.6.0.js"' in result
        assert 'src="/static/tooltipster.bundle.js"' in result

    def test_standalone_true_uses_cdn_for_jquery(self):
        """Test that standalone=True uses CDN URL for jQuery."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Should use jsdelivr CDN for jQuery 3.6.0
        assert "cdn.jsdelivr.net/npm/jquery@3.6.0" in result
        # Should NOT use /static/ path
        assert 'src="/static/jquery-3.6.0.js"' not in result

    def test_standalone_true_uses_cdn_for_tooltipster(self):
        """Test that standalone=True uses CDN URL for Tooltipster."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Should use CDN for tooltipster CSS and JS
        assert "cdn.jsdelivr.net/npm/tooltipster@4" in result
        # Should NOT use /static/ paths
        assert 'href="/static/tooltipster.bundle.css"' not in result
        assert 'src="/static/tooltipster.bundle.js"' not in result

    def test_standalone_true_uses_studio_url_for_rsm_assets(self):
        """Test that standalone=True uses Studio backend URL for RSM assets."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # RSM CSS should point to Studio backend (or be inlined)
        # The exact URL format will be determined by implementation
        assert 'href="/static/rsm.css"' not in result
        # Should have some form of rsm.css reference (either CDN or inline)
        assert "rsm.css" in result or "<style" in result

    def test_standalone_true_uses_cdn_for_mathjax(self):
        """Test that standalone=True preserves MathJax CDN URL."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # MathJax should still use CDN (it already does)
        # This test ensures we don't accidentally break it
        assert "mathjax" in result.lower() or "cdn.jsdelivr.net" in result

    def test_standalone_true_uses_cdn_for_pseudocode(self):
        """Test that standalone=True preserves pseudocode.js CDN URL."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Pseudocode CSS should use CDN
        assert "pseudocode" in result.lower()

    def test_standalone_output_is_valid_html(self):
        """Test that standalone output is a complete, valid HTML document."""
        source = ":rsm:\n\n# Test Title\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Must have all required HTML structure
        assert "<html" in result.lower()
        assert "<head" in result.lower()
        assert "<body" in result.lower()
        assert "</html>" in result.lower()
        assert "<title>" in result.lower()

    def test_standalone_default_is_false(self):
        """Test that standalone defaults to False for backwards compatibility."""
        source = ":rsm:\n\nHello world.\n\n::"
        result = rsm.build(source, handrails=False, lint=False)

        # Default behavior should use /static/ paths
        assert 'href="/static/rsm.css"' in result


class TestStandaloneBuilder:
    """Test the StandaloneBuilder class directly."""

    def test_standalone_builder_exists(self):
        """Test that StandaloneBuilder class exists."""
        from rsm.builder import StandaloneBuilder

        assert StandaloneBuilder is not None

    def test_standalone_builder_inherits_from_html_builder(self):
        """Test that StandaloneBuilder inherits from HTMLBuilder."""
        from rsm.builder import HTMLBuilder, StandaloneBuilder

        assert issubclass(StandaloneBuilder, HTMLBuilder)

    def test_standalone_builder_produces_single_file(self):
        """Test that StandaloneBuilder produces exactly one file."""
        from rsm.builder import StandaloneBuilder

        standalone_builder = StandaloneBuilder()
        body = '<body><div class="manuscript">Test</div></body>'
        web = standalone_builder.build(body)

        # Should only create index.html, no static directory
        files = list(web.listdir("/"))
        assert "index.html" in files
        assert "static" not in files


class TestStandaloneCDNVersions:
    """Test that CDN URLs use correct library versions."""

    def test_jquery_version_is_3_6_0(self):
        """Test jQuery CDN uses exact version 3.6.0 to match bundled version."""
        source = ":rsm:\n\nTest.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Must be exact version, not "latest"
        assert "jquery@3.6.0" in result

    def test_tooltipster_version_matches_bundled(self):
        """Test Tooltipster CDN uses version compatible with bundled version."""
        source = ":rsm:\n\nTest.\n\n::"
        result = rsm.build(source, handrails=False, lint=False, standalone=True)

        # Tooltipster 4.x series
        assert "tooltipster@4" in result


class TestCustomCSS:
    """Test custom CSS support in builders."""

    def test_folder_builder_with_custom_css(self, tmp_path):
        """Test FolderBuilder copies custom CSS to static/ folder."""
        from rsm.builder import FolderBuilder

        # Create custom CSS file
        custom_css_file = tmp_path / "custom.css"
        custom_css_content = ".mytype { color: red; }"
        custom_css_file.write_text(custom_css_content)

        # Create builder with custom_css
        builder = FolderBuilder(custom_css=str(custom_css_file))
        body = '<body><div class="manuscript">Test</div></body>'
        web = builder.build(body, src=tmp_path / "test.rsm")

        # Should copy custom.css to static/ folder
        assert web.exists("static/custom.css")
        custom_css_in_static = web.readtext("static/custom.css")
        assert custom_css_content in custom_css_in_static

        # Should add link tag in HTML
        html = web.readtext("index.html")
        assert '<link rel="stylesheet" type="text/css" href="/static/custom.css"' in html

    def test_standalone_builder_with_custom_css(self, tmp_path):
        """Test StandaloneBuilder inlines custom CSS in <style> tag."""
        from rsm.builder import StandaloneBuilder

        # Create custom CSS file
        custom_css_file = tmp_path / "custom.css"
        custom_css_content = ".mytype { color: blue; }"
        custom_css_file.write_text(custom_css_content)

        # Create builder with custom_css
        builder = StandaloneBuilder(custom_css=str(custom_css_file))
        body = '<body><div class="manuscript">Test</div></body>'
        web = builder.build(body, src=tmp_path / "test.rsm")

        # Should inline custom CSS in <style> tag
        html = web.readtext("index.html")
        assert "<style>" in html
        assert custom_css_content in html

        # Should NOT create static/ folder
        files = list(web.listdir("/"))
        assert "static" not in files

    def test_builder_without_custom_css(self, tmp_path):
        """Test builders work normally when custom_css=None."""
        from rsm.builder import FolderBuilder

        # Create builder WITHOUT custom_css
        builder = FolderBuilder()
        body = '<body><div class="manuscript">Test</div></body>'
        web = builder.build(body, src=tmp_path / "test.rsm")

        # Should work normally
        assert web.exists("index.html")
        html = web.readtext("index.html")
        assert "Test" in html

        # Should NOT have custom.css link
        assert "custom.css" not in html

    def test_builder_custom_css_nonexistent_file(self, tmp_path):
        """Test builder handles nonexistent custom CSS gracefully."""
        from rsm.builder import FolderBuilder

        # Create builder with nonexistent CSS file
        builder = FolderBuilder(custom_css="nonexistent.css")
        body = '<body><div class="manuscript">Test</div></body>'
        web = builder.build(body, src=tmp_path / "test.rsm")

        # Should still build successfully
        assert web.exists("index.html")

        # Should NOT have custom.css in static/
        assert not web.exists("static/nonexistent.css")



"""Tests for rsm build CLI file output behavior.

The rsm build command should write files to disk (index.html + static/ folder).
The rsm.build() Python API should return HTML without writing files.
"""

import subprocess

import pytest

import rsm


class TestMakeCLIFileOutput:
    """Test that rsm build CLI writes files to disk."""

    @pytest.mark.slow
    def test_make_cli_creates_index_html(self, tmp_path):
        """Test that rsm build creates index.html file."""
        src = "# Test\n\nHello world.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        # Run rsm build in tmp_path directory
        subprocess.run(
            f"rsm build {src_file}",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create test.html (derived from test.rsm)
        output_html = tmp_path / "test.html"
        assert output_html.exists(), "rsm build should create test.html from test.rsm"
        assert output_html.read_text().strip() != "", "test.html should not be empty"
        assert "<html>" in output_html.read_text()
        assert "Hello world" in output_html.read_text()

    @pytest.mark.slow
    def test_make_cli_creates_static_folder(self, tmp_path):
        """Test that rsm build creates static/ folder with assets."""
        src = "# Test\n\nHello world.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        subprocess.run(
            f"rsm build {src_file}",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create static/ folder
        static_dir = tmp_path / "static"
        assert static_dir.exists(), "rsm build should create static/ folder"
        assert static_dir.is_dir()

        # Should contain CSS and JS files
        assert (static_dir / "rsm.css").exists()
        assert (static_dir / "jquery-3.6.0.js").exists()
        assert (static_dir / "tooltipster.bundle.js").exists()

    @pytest.mark.slow
    def test_make_cli_with_string_flag_creates_files_in_cwd(self, tmp_path):
        """Test that rsm build with -c flag creates files in current directory."""
        src = ":rsm:\n# Test\n\nString source.\n"

        # Run with -c flag in tmp_path
        subprocess.run(
            f'rsm build "{src}" -c',
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should still create files even with -c flag
        index_html = tmp_path / "index.html"
        assert index_html.exists()
        assert "String source" in index_html.read_text()

    @pytest.mark.slow
    def test_make_cli_default_no_stdout(self, tmp_path):
        """Test that rsm build default behavior produces no stdout."""
        src = "# Test\n\nDefault mode.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        result = subprocess.run(
            f"rsm build {src_file}",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should produce no stdout by default
        # Filter out warnings
        stdout = result.stdout.decode("utf-8")
        lines = [line for line in stdout.split("\n") if "pkg_resources" not in line]
        clean_stdout = "\n".join(lines).strip()
        assert clean_stdout == "", "Default mode should produce no output to stdout"

        # But should still create files (test.html from test.rsm)
        output_html = tmp_path / "test.html"
        assert output_html.exists()
        assert "Default mode" in output_html.read_text()

    @pytest.mark.slow
    def test_make_cli_output_flag_filename(self, tmp_path):
        """Test rsm build -o filename creates custom-named file."""
        src = "# Test\n\nCustom filename.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        subprocess.run(
            f"rsm build {src_file} -o myfile",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create myfile.html instead of index.html
        output_html = tmp_path / "myfile.html"
        assert output_html.exists()
        assert "Custom filename" in output_html.read_text()
        assert not (tmp_path / "index.html").exists()

    @pytest.mark.slow
    def test_make_cli_output_flag_directory(self, tmp_path):
        """Test rsm build -o dir/ creates files in directory."""
        src = "# Test\n\nCustom directory.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        subprocess.run(
            f"rsm build {src_file} -o build/",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create build/test.html (derived from test.rsm)
        build_dir = tmp_path / "build"
        assert build_dir.exists() and build_dir.is_dir()
        output_html = build_dir / "test.html"
        assert output_html.exists()
        assert "Custom directory" in output_html.read_text()

    @pytest.mark.slow
    def test_make_cli_output_flag_both(self, tmp_path):
        """Test rsm build -o dir/filename creates custom file in directory."""
        src = "# Test\n\nBoth custom.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        subprocess.run(
            f"rsm build {src_file} -o dist/document",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create dist/document.html
        dist_dir = tmp_path / "dist"
        assert dist_dir.exists() and dist_dir.is_dir()
        output_html = dist_dir / "document.html"
        assert output_html.exists()
        assert "Both custom" in output_html.read_text()

    @pytest.mark.slow
    def test_make_cli_print_flag(self, tmp_path):
        """Test rsm build -p prints HTML to stdout AND creates files."""
        src = "# Test\n\nPrint flag.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        result = subprocess.run(
            f"rsm build {src_file} -p",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should print HTML to stdout
        stdout = result.stdout.decode("utf-8")
        assert "<html>" in stdout
        assert "Print flag" in stdout

        # Should also create files (test.html from test.rsm)
        output_html = tmp_path / "test.html"
        assert output_html.exists()
        assert "Print flag" in output_html.read_text()

    @pytest.mark.slow
    def test_make_cli_standalone_flag(self, tmp_path):
        """Test rsm build --standalone creates single HTML file."""
        src = "# Test\n\nStandalone.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        subprocess.run(
            f"rsm build {src_file} --standalone",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should create test.html
        output_html = tmp_path / "test.html"
        assert output_html.exists()
        assert "Standalone" in output_html.read_text()

        # Should NOT create static/ folder
        static_dir = tmp_path / "static"
        assert not static_dir.exists()

        # Should use CDN URLs
        html_content = output_html.read_text()
        assert "cdn.jsdelivr.net" in html_content

    @pytest.mark.slow
    def test_make_cli_custom_css_flag(self, tmp_path):
        """Test rsm build --css copies custom CSS to static/ folder."""
        src = "# Test\n\n:paragraph: {:types: mytype} Custom styled text\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        # Create custom CSS file
        custom_css = ".mytype { color: red; }"
        custom_css_file = tmp_path / "custom.css"
        custom_css_file.write_text(custom_css)

        subprocess.run(
            f"rsm build {src_file} --css {custom_css_file}",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should copy custom.css to static/ folder
        static_dir = tmp_path / "static"
        custom_css_in_static = static_dir / "custom.css"
        assert custom_css_in_static.exists(), "Custom CSS should be copied to static/"
        assert custom_css in custom_css_in_static.read_text()

        # Should add link tag in HTML
        output_html = tmp_path / "test.html"
        html_content = output_html.read_text()
        assert '<link rel="stylesheet" type="text/css" href="/static/custom.css"' in html_content

        # Custom CSS link should come after rsm.css
        rsm_css_pos = html_content.find('/static/rsm.css')
        custom_css_pos = html_content.find('/static/custom.css')
        assert rsm_css_pos < custom_css_pos, "Custom CSS should load after rsm.css"

    @pytest.mark.slow
    def test_custom_css_overrides_rsm_css(self, tmp_path):
        """Test that rsm.css is wrapped in @layer base to allow custom CSS to override."""
        src = "# Test\n\nA simple paragraph.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        # Create custom CSS with low specificity (will still win due to @layer)
        custom_css = ".custom { color: red; }"
        custom_css_file = tmp_path / "custom.css"
        custom_css_file.write_text(custom_css)

        subprocess.run(
            f"rsm build {src_file} --css {custom_css_file}",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Verify custom CSS is present
        custom_css_in_static = tmp_path / "static" / "custom.css"
        assert custom_css in custom_css_in_static.read_text()

        # Verify rsm.css is wrapped in @layer base
        rsm_css_in_static = tmp_path / "static" / "rsm.css"
        rsm_css_content = rsm_css_in_static.read_text()
        assert "@layer base" in rsm_css_content, "rsm.css should use @layer base"

    @pytest.mark.slow
    def test_make_cli_custom_css_standalone(self, tmp_path):
        """Test rsm build --css --standalone inlines custom CSS."""
        src = "# Test\n\n:paragraph: {:types: mytype} Custom styled text\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        # Create custom CSS file
        custom_css = ".mytype { color: red; }"
        custom_css_file = tmp_path / "custom.css"
        custom_css_file.write_text(custom_css)

        subprocess.run(
            f"rsm build {src_file} --css {custom_css_file} --standalone",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
            check=True,
        )

        # Should NOT create static/ folder
        static_dir = tmp_path / "static"
        assert not static_dir.exists(), "--standalone should not create static/ folder"

        # Should inline custom CSS in <style> tag
        output_html = tmp_path / "test.html"
        html_content = output_html.read_text()
        assert "<style>" in html_content
        assert custom_css in html_content

        # <style> tag should come after CDN rsm.css link
        cdn_css_pos = html_content.find('cdn.jsdelivr.net')
        style_tag_pos = html_content.find('<style>')
        assert cdn_css_pos < style_tag_pos, "Inline custom CSS should come after CDN rsm.css"

    @pytest.mark.slow
    def test_make_cli_custom_css_nonexistent_file(self, tmp_path):
        """Test rsm build with nonexistent CSS file logs warning but continues."""
        src = "# Test\n\nNonexistent CSS.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        result = subprocess.run(
            f"rsm build {src_file} --css nonexistent.css",
            cwd=tmp_path,
            shell=True,
            capture_output=True,
        )

        # Should still succeed (or at least not crash)
        # Check stderr for warning
        stderr = result.stderr.decode("utf-8")
        assert "nonexistent.css" in stderr or "not found" in stderr.lower()

        # Should still create test.html
        output_html = tmp_path / "test.html"
        assert output_html.exists()


class TestMakePythonAPINoFileOutput:
    """Test that rsm.build() Python API returns HTML without writing files."""

    def test_make_api_returns_html_string(self):
        """Test that rsm.build() returns HTML as string."""
        src = "# Test\n\nAPI call.\n"
        result = rsm.build(src)

        assert isinstance(result, str)
        assert "<html>" in result
        assert "API call" in result

    def test_make_api_does_not_write_files(self, tmp_path):
        """Test that rsm.build() does NOT write files to disk."""
        src = "# Test\n\nAPI call.\n"

        # Change to tmp_path to ensure no files are written
        import os

        original_cwd = os.getcwd()
        try:
            os.chdir(tmp_path)

            # Call API
            result = rsm.build(src)
            assert isinstance(result, str)

            # Should NOT create index.html or static/
            assert not (tmp_path / "index.html").exists()
            assert not (tmp_path / "static").exists()

        finally:
            os.chdir(original_cwd)

    def test_make_api_with_path_does_not_write_files(self, tmp_path):
        """Test that rsm.build(path=...) does NOT write files to disk."""
        src = "# Test\n\nAPI with path.\n"
        src_file = tmp_path / "test.rsm"
        src_file.write_text(src)

        # Call API with path
        result = rsm.build(path=str(src_file))
        assert isinstance(result, str)
        assert "API with path" in result

        # Should NOT create index.html or static/ in tmp_path
        assert not (tmp_path / "index.html").exists()
        assert not (tmp_path / "static").exists()



from conftest import compare_have_want


def test_simple_no_handrails():
    compare_have_want(
        have="""\
        :sketch: Foo.::

        :proof:

          :step: Bar.::

        ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="sketch" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof sketch.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>Foo.</p>

        </div>

        </div>

        <div class="proof" data-nodeid="4">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof.</span></p>

        </div>

        <div class="step last" data-nodeid="5">

        <div class="statement" data-nodeid="6">

        <div class="paragraph" data-nodeid="7">

        <p>Bar.</p>

        </div>

        </div>

        </div>

        <div class="halmos"></div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""\
        ## Section

        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_subsections():
    compare_have_want(
        have="""\
        ## Section

        Lorem ipsum.

        ### Sub section

        Foo

        #### Sub sub section

        Bar

        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Lorem ipsum.</p>

        </div>

        <section class="subsection level-3" data-nodeid="4">

        <h3>1.1. Sub section</h3>

        <div class="paragraph" data-nodeid="5">

        <p>Foo</p>

        </div>

        <section class="subsubsection level-4" data-nodeid="7">

        <h4>1.1.1. Sub sub section</h4>

        <div class="paragraph" data-nodeid="8">

        <p>Bar</p>

        </div>

        </section>

        </section>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_code():
    compare_have_want(
        have="""\
        :code: comp = [abs(x) for x in range(10)] ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="code" data-nodeid="2"><code>comp = [abs(x) for x in range(10)]</code></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_code_with_shorthand():
    compare_have_want(
        have="""\
        `comp = [abs(x) for x in range(10)]`
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="code" data-nodeid="2"><code>comp = [abs(x) for x in range(10)]</code></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_code_with_lang():
    compare_have_want(
        have="""\
        :code: {:lang: python}
          comp = [abs(x) for x in range(10)]
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="code" data-nodeid="2"><code class="highlight python"><span class="n">comp</span> <span class="o">=</span> <span class="p">[</span><span class="nb">abs</span><span class="p">(</span><span class="n">x</span><span class="p">)</span> <span class="k">for</span> <span class="n">x</span> <span class="ow">in</span> <span class="nb">range</span><span class="p">(</span><span class="mi">10</span><span class="p">)]</span>
        </code></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_code_with_shorthand_and_lang():
    compare_have_want(
        have="""\
        `{:lang: python} comp = [abs(x) for x in range(10)]`
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="code" data-nodeid="2"><code class="highlight python"><span class="n">comp</span> <span class="o">=</span> <span class="p">[</span><span class="nb">abs</span><span class="p">(</span><span class="n">x</span><span class="p">)</span> <span class="k">for</span> <span class="n">x</span> <span class="ow">in</span> <span class="nb">range</span><span class="p">(</span><span class="mi">10</span><span class="p">)]</span>
        </code></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_codeblock():
    compare_have_want(
        have="""\
        :codeblock:

          comp = [abs(x) for x in range(10)]

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>comp = [abs(x) for x in range(10)]</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_codeblock_with_lang():
    compare_have_want(
        have="""\
        :codeblock: {:lang: python}

          comp = [abs(x) for x in range(10)]

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code class="highlight python"><span class="n">comp</span> <span class="o">=</span> <span class="p">[</span><span class="nb">abs</span><span class="p">(</span><span class="n">x</span><span class="p">)</span> <span class="k">for</span> <span class="n">x</span> <span class="ow">in</span> <span class="nb">range</span><span class="p">(</span><span class="mi">10</span><span class="p">)]</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_codeblock_with_shorthand():
    compare_have_want(
        have="""\
        ```
        comp = [abs(x) for x in range(10)]
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>comp = [abs(x) for x in range(10)]</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_codeblock_with_shorthand_and_lang():
    compare_have_want(
        have="""\
        ```{:lang: python}

        comp = [abs(x) for x in range(10)]

        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code class="highlight python"><span class="n">comp</span> <span class="o">=</span> <span class="p">[</span><span class="nb">abs</span><span class="p">(</span><span class="n">x</span><span class="p">)</span> <span class="k">for</span> <span class="n">x</span> <span class="ow">in</span> <span class="nb">range</span><span class="p">(</span><span class="mi">10</span><span class="p">)]</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_meta():
    compare_have_want(
        have="""\
        `{:label: 1}k=v`

        :code:{:label: 2}k=v::

        :codeblock: {
          :label: 3
        }
          k = v
        ::

        ```
          {:label: 4}
          k = v
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span id="1" class="code" data-nodeid="2"><code>k=v</code></span></p>

        </div>

        <div class="paragraph" data-nodeid="4">

        <p><span id="2" class="code" data-nodeid="5"><code>k=v</code></span></p>

        </div>

        <div id="3" class="codeblock" data-nodeid="7">

        <pre>
        <code>k = v</code>
        </pre>

        </div>

        <div id="4" class="codeblock" data-nodeid="9">

        <pre>
        <code>k = v</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_meta_and_lang():
    compare_have_want(
        have="""\
        `{:label: 1, :lang: js}k=v`

        :code:{:label: 2, :lang: js}k=v::

        :codeblock: {
          :label: 3
          :lang: js
        }
          k = v
        ::

        ``` {
          :label: 4
          :lang: js
        }
          k = v
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span id="1" class="code" data-nodeid="2"><code class="highlight js"><span class="nx">k</span><span class="o">=</span><span class="nx">v</span>
        </code></span></p>

        </div>

        <div class="paragraph" data-nodeid="4">

        <p><span id="2" class="code" data-nodeid="5"><code class="highlight js"><span class="nx">k</span><span class="o">=</span><span class="nx">v</span>
        </code></span></p>

        </div>

        <div id="3" class="codeblock" data-nodeid="7">

        <pre>
        <code class="highlight js"><span class="nx">k</span><span class="w"> </span><span class="o">=</span><span class="w"> </span><span class="nx">v</span>
        </code>
        </pre>

        </div>

        <div id="4" class="codeblock" data-nodeid="9">

        <pre>
        <code class="highlight js"><span class="nx">k</span><span class="w"> </span><span class="o">=</span><span class="w"> </span><span class="nx">v</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_codeblock():
    compare_have_want(
        have="""\
        :codeblock:

        import sys
        print(sys.argv)

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>import sys
        print(sys.argv)</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_codeblock_with_lang():
    compare_have_want(
        have="""\
        :codeblock:{:lang: python}

        import sys
        print(sys.argv)

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code class="highlight python"><span class="kn">import</span><span class="w"> </span><span class="nn">sys</span>
        <span class="nb">print</span><span class="p">(</span><span class="n">sys</span><span class="o">.</span><span class="n">argv</span><span class="p">)</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_codeblock_with_shorthand():
    compare_have_want(
        have="""\
        ```
        import sys
        print(sys.argv)
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>import sys
        print(sys.argv)</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_codeblock_with_shorthand_and_lang():
    compare_have_want(
        have="""\
        ```
        {:lang: python}

        import sys
        print(sys.argv)

        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code class="highlight python"><span class="kn">import</span><span class="w"> </span><span class="nn">sys</span>
        <span class="nb">print</span><span class="p">(</span><span class="n">sys</span><span class="o">.</span><span class="n">argv</span><span class="p">)</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_meta():
    compare_have_want(
        have="""\
        :codeblock:{:label: 3}
        k = v
        key = val
        ::

        ```
        {:label: 4}
        k = v
        key = val
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div id="3" class="codeblock" data-nodeid="1">

        <pre>
        <code>k = v
        key = val</code>
        </pre>

        </div>

        <div id="4" class="codeblock" data-nodeid="3">

        <pre>
        <code>k = v
        key = val</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiline_meta_and_lang():
    compare_have_want(
        have="""\
        :codeblock: {
          :label: 3
          :lang: lisp
        }
        (defun forall (list func)
         (if (null list)
           t
           (and (funcall func (car list))
           (forall (cdr list) func))))
        ::

        ``` {
          :label: 4
          :lang: lisp
        }
        (defun forall (list func)
         (if (null list)
           t
           (and (funcall func (car list))
           (forall (cdr list) func))))
        ```
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div id="3" class="codeblock" data-nodeid="1">

        <pre>
        <code class="highlight lisp"><span class="p">(</span><span class="nb">defun</span><span class="w"> </span><span class="nv">forall</span><span class="w"> </span><span class="p">(</span><span class="nb">list</span><span class="w"> </span><span class="nv">func</span><span class="p">)</span>
        <span class="w"> </span><span class="p">(</span><span class="k">if</span><span class="w"> </span><span class="p">(</span><span class="nb">null</span><span class="w"> </span><span class="nb">list</span><span class="p">)</span>
        <span class="w">   </span><span class="no">t</span>
        <span class="w">   </span><span class="p">(</span><span class="nb">and</span><span class="w"> </span><span class="p">(</span><span class="nb">funcall</span><span class="w"> </span><span class="nv">func</span><span class="w"> </span><span class="p">(</span><span class="nb">car</span><span class="w"> </span><span class="nb">list</span><span class="p">))</span>
        <span class="w">   </span><span class="p">(</span><span class="nv">forall</span><span class="w"> </span><span class="p">(</span><span class="nb">cdr</span><span class="w"> </span><span class="nb">list</span><span class="p">)</span><span class="w"> </span><span class="nv">func</span><span class="p">))))</span>
        </code>
        </pre>

        </div>

        <div id="4" class="codeblock" data-nodeid="3">

        <pre>
        <code class="highlight lisp"><span class="p">(</span><span class="nb">defun</span><span class="w"> </span><span class="nv">forall</span><span class="w"> </span><span class="p">(</span><span class="nb">list</span><span class="w"> </span><span class="nv">func</span><span class="p">)</span>
        <span class="w"> </span><span class="p">(</span><span class="k">if</span><span class="w"> </span><span class="p">(</span><span class="nb">null</span><span class="w"> </span><span class="nb">list</span><span class="p">)</span>
        <span class="w">   </span><span class="no">t</span>
        <span class="w">   </span><span class="p">(</span><span class="nb">and</span><span class="w"> </span><span class="p">(</span><span class="nb">funcall</span><span class="w"> </span><span class="nv">func</span><span class="w"> </span><span class="p">(</span><span class="nb">car</span><span class="w"> </span><span class="nb">list</span><span class="p">))</span>
        <span class="w">   </span><span class="p">(</span><span class="nv">forall</span><span class="w"> </span><span class="p">(</span><span class="nb">cdr</span><span class="w"> </span><span class="nb">list</span><span class="p">)</span><span class="w"> </span><span class="nv">func</span><span class="p">))))</span>
        </code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""Regression test for closure bug in rsm serve file watching.

Bug: In the file watching loop in _cmd_serve(), the `build_cmd` variable must be
captured by value (using a default argument) instead of by reference. Without this,
all callbacks share the same variable which ends up with the last iteration's value.

Example: If you have test1.rsm, test2.rsm, test3.rsm:
  - Buggy: Editing test1.rsm would trigger build of test3.rsm
  - Fixed: Editing test1.rsm triggers build of test1.rsm

The fix: Change `def rebuild_callback(rsm_path=rsm_file)`
       to `def rebuild_callback(rsm_path=rsm_file, cmd=build_cmd)`
"""


def test_closure_bug_demonstration():
    """Demonstrate how the closure bug manifests and how to fix it.

    This test documents the bug pattern and demonstrates both the buggy
    and fixed versions work.
    """
    # BUGGY VERSION: Variable captured by reference
    callbacks_buggy = []
    for i in range(3):
        build_cmd = f"command_{i}"
        # BUG: build_cmd is captured by reference, not value
        def callback():
            return build_cmd
        callbacks_buggy.append(callback)

    # All callbacks return the SAME value (from last iteration)
    assert callbacks_buggy[0]() == "command_2"  # Wrong! Should be "command_0"
    assert callbacks_buggy[1]() == "command_2"  # Wrong! Should be "command_1"
    assert callbacks_buggy[2]() == "command_2"  # Correct

    # FIXED VERSION: Variable captured by value using default argument
    callbacks_fixed = []
    for i in range(3):
        build_cmd = f"command_{i}"
        # FIX: build_cmd is captured by value as a default argument
        def callback(cmd=build_cmd):
            return cmd
        callbacks_fixed.append(callback)

    # Each callback returns its OWN value
    assert callbacks_fixed[0]() == "command_0"  # Correct!
    assert callbacks_fixed[1]() == "command_1"  # Correct!
    assert callbacks_fixed[2]() == "command_2"  # Correct!


def test_rsm_serve_specific_scenario():
    """Demonstrate the exact scenario from rsm serve code."""
    from pathlib import Path

    # Simulate the loop in _cmd_serve() for multiple .rsm files
    rsm_files = [Path(f"test{i}.rsm") for i in range(1, 4)]

    # BUGGY VERSION (what happens without the fix)
    callbacks_buggy = {}
    for rsm_file in rsm_files:
        output_name = rsm_file.stem
        build_cmd = f"rsm build {rsm_file} -o {output_name}"
        # BUG: build_cmd not captured as default arg
        def rebuild_callback(rsm_path=rsm_file):
            return build_cmd
        callbacks_buggy[rsm_file.name] = rebuild_callback

    # All callbacks return the same command (the last one)!
    assert callbacks_buggy["test1.rsm"]() == "rsm build test3.rsm -o test3"
    assert callbacks_buggy["test2.rsm"]() == "rsm build test3.rsm -o test3"
    assert callbacks_buggy["test3.rsm"]() == "rsm build test3.rsm -o test3"

    # FIXED VERSION (with the fix applied)
    callbacks_fixed = {}
    for rsm_file in rsm_files:
        output_name = rsm_file.stem
        build_cmd = f"rsm build {rsm_file} -o {output_name}"
        # FIX: build_cmd captured as default arg
        def rebuild_callback(rsm_path=rsm_file, cmd=build_cmd):
            return cmd
        callbacks_fixed[rsm_file.name] = rebuild_callback

    # Each callback returns the correct command!
    assert callbacks_fixed["test1.rsm"]() == "rsm build test1.rsm -o test1"
    assert callbacks_fixed["test2.rsm"]() == "rsm build test2.rsm -o test2"
    assert callbacks_fixed["test3.rsm"]() == "rsm build test3.rsm -o test3"



from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""\
        :theorem:

        Theorem contents.

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="theorem" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>Theorem contents.</p>

        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_multiple():
    compare_have_want(
        have="""\
        :theorem:

        Theorem contents.

        ::

        :lemma:

        And a lemma.

        ::

        :theorem:

        Another theorem.

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="theorem" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>Theorem contents.</p>

        </div>

        </div>

        <div class="lemma theorem" data-nodeid="4">

        <div class="paragraph hr-label">

        <p><span class="span label">Lemma 2.</span></p>

        </div>

        <div class="paragraph" data-nodeid="5">

        <p>And a lemma.</p>

        </div>

        </div>

        <div class="theorem" data-nodeid="7">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 3.</span></p>

        </div>

        <div class="paragraph" data-nodeid="8">

        <p>Another theorem.</p>

        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""Test that multiple RSM files in the same directory handle assets correctly."""

import subprocess
from pathlib import Path


def test_multiple_files_with_different_images(tmp_path):
    """Test that multiple RSM files can reference different images.

    The images are NOT copied to static/ - they're referenced with relative
    paths from the RSM source. Only the RSM framework files (CSS, JS) go in
    static/, and those are identical for all documents.
    """
    # Create images directory
    images_dir = tmp_path / "images"
    images_dir.mkdir()

    # Create placeholder images
    (images_dir / "pic1.png").write_bytes(b"fake-png-1")
    (images_dir / "pic2.png").write_bytes(b"fake-png-2")

    # Create doc1.rsm that references pic1.png
    doc1 = tmp_path / "doc1.rsm"
    doc1.write_text("""\
# Document 1

:figure: {:path: images/pic1.png}
:caption: Picture 1
::
""")

    # Create doc2.rsm that references pic2.png
    doc2 = tmp_path / "doc2.rsm"
    doc2.write_text("""\
# Document 2

:figure: {:path: images/pic2.png}
:caption: Picture 2
::
""")

    # Build both documents
    subprocess.run(["rsm", "build", "doc1.rsm"], cwd=tmp_path, check=True, capture_output=True)
    subprocess.run(["rsm", "build", "doc2.rsm"], cwd=tmp_path, check=True, capture_output=True)

    # Check outputs exist
    html1 = tmp_path / "doc1.html"
    html2 = tmp_path / "doc2.html"
    assert html1.exists()
    assert html2.exists()

    # Check that each HTML references the correct image
    content1 = html1.read_text()
    content2 = html2.read_text()

    # Check that image paths are in the HTML (checking for presence, not exact format)
    assert 'images/pic1.png' in content1, "doc1.html should reference pic1.png"
    assert 'images/pic2.png' in content2, "doc2.html should reference pic2.png"
    assert '<img src=' in content1, "doc1.html should have an img tag"
    assert '<img src=' in content2, "doc2.html should have an img tag"

    # Images should NOT be copied to static/
    static_dir = tmp_path / "static"
    assert not (static_dir / "pic1.png").exists(), "Images should NOT be copied to static/"
    assert not (static_dir / "pic2.png").exists(), "Images should NOT be copied to static/"
    assert not (static_dir / "images").exists(), "Images directory should NOT be in static/"

    # Images should remain in their original location
    assert (images_dir / "pic1.png").exists(), "Original images should remain in place"
    assert (images_dir / "pic2.png").exists(), "Original images should remain in place"

    # Both documents share the same static/ folder for RSM framework files
    assert (static_dir / "rsm.css").exists(), "Both should share RSM CSS in static/"


def test_multiple_files_share_static_assets(tmp_path):
    """Test that multiple RSM files share the same static/ directory.

    This is correct behavior because the static/ directory only contains
    RSM framework files (CSS, JS) which are identical for all documents.
    """
    # Create three RSM files
    for i in range(1, 4):
        rsm_file = tmp_path / f"doc{i}.rsm"
        rsm_file.write_text(f"# Document {i}\n\nContent {i}.")
        subprocess.run(["rsm", "build", f"doc{i}.rsm"], cwd=tmp_path, check=True, capture_output=True)

    # All three should create HTML files
    assert (tmp_path / "doc1.html").exists()
    assert (tmp_path / "doc2.html").exists()
    assert (tmp_path / "doc3.html").exists()

    # All three share the same static/ directory
    static_dir = tmp_path / "static"
    assert static_dir.exists()

    # All HTML files reference the same static assets
    for i in range(1, 4):
        html_file = tmp_path / f"doc{i}.html"
        content = html_file.read_text()
        assert '/static/rsm.css' in content, f"doc{i}.html should reference shared static/rsm.css"

    # The static files are the same for all documents, so sharing is safe
    rsm_css = (static_dir / "rsm.css").read_text()
    assert len(rsm_css) > 0, "Static CSS should exist and be non-empty"



"""Applications that execute the steps of the RSM file processing pipeline.

Classes in this module use the steps implemented elsewhere (:mod:`rsm.reader`,
:mod:`rsm.tsparser`, etc) and actually execute those steps, feeding the output of one
step into the next.  The base class for all of these is :class:`RSMApp`, which stores
the sequence of steps and executes them all via a ``run()`` method.

The three main applications provided currently are :func:`make`, :func:`render`, and
:func:`lint`.  These simply instantiate an object of class :class:`FullBuildApp`,
:class:`LinterApp`, :class:`ProcessorApp`, respectively, and call their ``run()``
method.

The functions :func:`make`, :func:`render`, and :func:`lint` can be called directly via
``rsm.make()``, ``rsm.render()``, and ``rsm.lint()``.  They each receive RSM source
(either a string or a path to a file), run the application, and return the result.  For
more control over the execution, or post-execution inspection, instantiate the
appropriate class manually.  For example:

.. code-block:: python

   This
   >>> src = "Hello, RSM!"
   >>> html = rsm.build(source=src)

   Is essentially equivalent to the following.

   >>> app = FullBuildApp(plain=src)
   >>> html = app.run()

   Except that now, ``app`` can be inspected.

   >>> print(app.parser.ast.sexp())
   (Manuscript
     (Paragraph
       (Text)))

The module :mod:`rsm.cli` exposes these apps as command line utilities.

The RSM library code (i.e. all of the previous modules) does not itself set up any
logging facilities.  This is at `recommendation
<https://docs.python.org/3/howto/logging-cookbook.html#adding-handlers-other-than-nullhandler-to-a-logger-in-a-library>`_
of the Python standard library.  Instead, this module uses the functions in
:mod:`rsm.rsmlogger` before executing an application.  In other words, manually handling
RSM source with classes such as :class:`rsm.tsparser.TSParser` or
:class:`rsm.transformer.Transformer` will not issue any logging messages, while calling
the ``run()`` method on an :class:`RSMApp` instance will.

"""

import logging
from collections.abc import Callable
from pathlib import Path
from typing import Any, NamedTuple

from rsm import (
    builder,
    linter,
    reader,
    rsmlogger,
    transformer,
    translator,
    tsparser,
    writer,
)
from rsm.tsparser import RSMParserError

from .rsmlogger import GatherHandler

logger = logging.getLogger("RSM")


def _parse_html_to_structured(html: str) -> dict:
    """Parse full HTML document into structured components.

    Parameters
    ----------
    html : str
        Complete HTML document with <html><head>...</head><body>...</body></html>

    Returns
    -------
    dict
        Dictionary with keys 'head', 'body', 'init_script'
    """
    import re

    # Extract head content (everything inside <head>...</head>)
    head_match = re.search(r"<head>(.*?)</head>", html, re.DOTALL)
    if not head_match:
        raise RSMApplicationError("No <head> section found in HTML document")

    head_content = head_match.group(1).strip()

    # Extract body content (everything inside <body>...</body>)
    # Use greedy .* to match the LAST </body> tag, not the first
    # (embedded HTML assets like Plotly charts contain their own </body> tags)
    body_match = re.search(r"<body[^>]*>(.*)</body>", html, re.DOTALL)
    if not body_match:
        raise RSMApplicationError("No <body> section found in HTML document")

    body_content = body_match.group(1).strip()

    # Extract external script tags from body content and add to head
    # This handles scripts added by HTML assets (like Plotly CDN)
    external_scripts = []
    script_pattern = r'<script[^>]*src=["\']([^"\']+)["\'][^>]*></script>'
    re.findall(script_pattern, body_content)

    for script_match in re.finditer(script_pattern, body_content):
        script_tag = script_match.group(0)
        src = script_match.group(1)

        # Only move external scripts (CDN URLs) to head
        if src.startswith(("http://", "https://", "//")):
            external_scripts.append(script_tag)
            # Remove from body content
            body_content = body_content.replace(script_tag, "")

    # Extract execution scripts from after </html> (created by _process_html_with_scripts)
    asset_init_scripts = []
    html_end = html.find("</html>")
    if html_end != -1:
        post_html_content = html[html_end:]

        # Extract CDN URLs for head injection
        src_pattern = r"script_[^.]+\.src\s*=\s*['\"]([^'\"]+)['\"]"
        execution_script_urls = re.findall(src_pattern, post_html_content)

        for url in execution_script_urls:
            if url.startswith(("http://", "https://", "//")):
                script_tag = f'<script src="{url}"></script>'
                external_scripts.append(script_tag)

        # Extract the full IIFE content for execution
        # The IIFE is wrapped in <script>...</script> after </html>
        script_content_pattern = r"<script>\s*(.*?)\s*</script>"
        script_matches = re.findall(
            script_content_pattern, post_html_content, re.DOTALL
        )
        for script_content in script_matches:
            if script_content.strip():
                asset_init_scripts.append(script_content.strip())

    # Add external scripts to head content
    if external_scripts:
        head_content = head_content + "\n" + "\n".join(external_scripts)

    # Clean up any extra whitespace in body content
    body_content = body_content.strip()

    # Combine asset scripts into init_script for frontend execution
    init_script = "\n".join(asset_init_scripts) if asset_init_scripts else ""

    return {"head": head_content, "body": body_content, "init_script": init_script}


class RSMApplicationError(Exception):
    pass


class Task(NamedTuple):
    """A step in a :class:`Pipeline`."""

    name: str
    obj: Any
    run: Callable


class Pipeline:
    """A sequence of :class:`Task` instances executed one after the other."""

    def __init__(self, tasks: list[Task]):
        self.tasks: list[Task] = []
        for t in tasks:
            self.add_task(t)

    def add_task(self, task: Task) -> None:
        """Add a task at the end of the current pipeline."""
        self.tasks.append(task)
        setattr(self, task.name, task.obj)

    def pop_task(self) -> Task:
        """Remove and return the last task."""
        task = self.tasks.pop()
        delattr(self, task.name)
        return task

    def run(self, initial_args: Any = None) -> Any:
        """Execute every task in the pipeline serially."""
        res = initial_args
        for _, _, call in self.tasks:
            if isinstance(res, dict):
                res = call(**res)
            elif isinstance(res, list | tuple):
                res = call(*res)
            elif res is None:
                res = call()
            else:
                res = call(res)
        return res


class RSMApp(Pipeline):
    default_log_level = logging.WARNING

    def __init__(
        self,
        tasks: list[Task] | None = None,
        loglevel: int = default_log_level,
        log_format: str = "rsm",
        log_time: bool = True,
        log_lineno: bool = True,
    ):
        rsmlogger.config_rsm_logger(loglevel, log_format, log_time, log_lineno)
        logger.info("Application started")
        logger.info("Configuring...")
        # self.config = self.config.configure()
        super().__init__(tasks or [])

    def run(self, initial_args: Any = None) -> str | None:
        try:
            result = super().run(initial_args)
            logger.info("Done.")
        except RSMParserError as e:
            logger.error(f"There was an error: {e}")
            result = ""
        return result

    def _setup_handler(self) -> None:
        target = logging.StreamHandler()
        target.setLevel(logging.WARNING)
        handler = GatherHandler([logging.WARNING], target)
        logger.addHandler(handler)

    @property
    def logs(self) -> list:
        return []


class ParserApp(RSMApp):
    def __init__(
        self,
        srcpath: Path | None = None,
        plain: str = "",
        loglevel: int = RSMApp.default_log_level,
        log_format: str = "rsm",
        log_time: bool = True,
        log_lineno: bool = True,
    ):
        self._validate_srcpath_and_plain(srcpath, plain)

        tasks = []
        if not plain:
            tasks.append(Task("reader", r := reader.Reader(), lambda: r.read(srcpath)))
        else:
            tasks.append(Task("dummy", None, lambda: plain))

        tasks += [
            Task("parser", p := tsparser.TSParser(), p.parse),
            Task("transformer", t := transformer.Transformer(), t.transform),
        ]
        super().__init__(tasks, loglevel, log_format, log_time, log_lineno)

    @staticmethod
    def _validate_srcpath_and_plain(srcpath: Path | str | None, plain: str) -> None:
        if (not srcpath and not plain) or (srcpath and plain):
            raise RSMApplicationError("Must specify exactly one of srcpath, plain")


class LinterApp(ParserApp):
    def __init__(
        self,
        srcpath: Path | None = None,
        plain: str = "",
        loglevel: int = RSMApp.default_log_level,
        log_format: str = "rsm",
        log_time: bool = True,
        log_lineno: bool = True,
    ):
        super().__init__(
            srcpath, plain, linter.Linter.LINT_LVL, log_format, log_time, log_lineno
        )
        mylinter = linter.Linter()
        self.add_task(Task("linter", mylinter, mylinter.lint))


class ProcessorApp(ParserApp):
    def __init__(
        self,
        srcpath: Path | None = None,
        plain: str = "",
        loglevel: int = RSMApp.default_log_level,
        log_format: str = "rsm",
        log_time: bool = True,
        log_lineno: bool = True,
        handrails: bool = False,
        add_source: bool = True,
        run_linter: bool = False,
        asset_resolver=None,
    ):
        super().__init__(srcpath, plain, loglevel, log_format, log_time, log_lineno)
        self.asset_resolver = asset_resolver
        if run_linter:
            self.add_task(
                Task("linter", linter_instance := linter.Linter(), linter_instance.lint)
            )

        if not handrails:
            tr = translator.Translator(asset_resolver=asset_resolver)
        else:
            tr = translator.HandrailsTranslator(
                add_source=add_source, asset_resolver=asset_resolver
            )
        self.add_task(Task("translator", tr, tr.translate))


class FullBuildApp(ProcessorApp):
    def __init__(
        self,
        srcpath: Path | None = None,
        plain: str = "",
        loglevel: int = RSMApp.default_log_level,
        log_format: str = "rsm",
        log_time: bool = True,
        log_lineno: bool = True,
        handrails: bool = True,
        add_source: bool = True,
        run_linter: bool = False,
        asset_resolver=None,
        standalone: bool = False,
        write_output: bool = False,
        output_dir: str = ".",
        output_filename: str = "index.html",
        custom_css: str | None = None,
    ):
        super().__init__(
            srcpath,
            plain,
            loglevel,
            log_format,
            log_time,
            log_lineno,
            handrails,
            add_source,
            run_linter,
            asset_resolver,
        )
        self.write_output = write_output
        if standalone:
            b = builder.StandaloneBuilder(asset_resolver=asset_resolver, outname=output_filename, custom_css=custom_css)
        else:
            b = builder.FolderBuilder(asset_resolver=asset_resolver, outname=output_filename, custom_css=custom_css)
        self.add_task(Task("builder", b, b.build))
        self.add_task(Task("writer", w := writer.Writer(dstpath=Path(output_dir)), w.write))

    def run(self, initial_args=None) -> str:
        """Run the build pipeline, optionally writing files to disk.

        If write_output=True, runs all tasks including writer and writes files.
        If write_output=False, skips writer and returns HTML content.
        """
        result = initial_args

        if self.write_output:
            # Run all tasks including writer
            for _, _, call in self.tasks:
                if isinstance(result, dict):
                    result = call(**result)
                elif isinstance(result, list | tuple):
                    result = call(*result)
                elif result is None:
                    result = call()
                else:
                    result = call(result)

            # Writer returns None, so we need to get HTML from the WebManuscript
            # The builder result is stored before writer runs
            for name, obj, _ in self.tasks:
                if name == "builder" and hasattr(obj, "web") and obj.web:
                    return obj.web.html

            logger.error("Builder did not produce WebManuscript")
            return ""
        else:
            # Run all tasks except writer
            for _, _, call in self.tasks[:-1]:
                if isinstance(result, dict):
                    result = call(**result)
                elif isinstance(result, list | tuple):
                    result = call(*result)
                elif result is None:
                    result = call()
                else:
                    result = call(result)

            # The result should now be a WebManuscript from the builder
            if hasattr(result, "html"):
                return result.html
            else:
                logger.error(
                    "Builder did not produce WebManuscript with html attribute"
                )
                return ""


def render(
    source: str = "",
    path: str = "",
    handrails: bool = False,
    add_source: bool = False,
    loglevel: int = RSMApp.default_log_level,
    log_format: str = "rsm",
    log_time: bool = True,
    log_lineno: bool = True,
    asset_resolver=None,
) -> str:
    return ProcessorApp(
        srcpath=path,
        plain=source,
        handrails=handrails,
        add_source=add_source,
        loglevel=loglevel,
        log_format=log_format,
        log_time=log_time,
        log_lineno=log_lineno,
        asset_resolver=asset_resolver,
    ).run()


def lint(
    source: str = "",
    path: str = "",
    handrails: bool = False,
    loglevel: int = RSMApp.default_log_level,
    log_format: str = "rsm",
    log_time: bool = True,
    log_lineno: bool = True,
):
    return LinterApp(
        srcpath=path,
        plain=source,
        loglevel=loglevel,
        log_format=log_format,
        log_time=log_time,
        log_lineno=log_lineno,
    ).run()


def build(
    source: str = "",
    path: str = "",
    handrails: bool = True,
    lint: bool = True,
    loglevel: int = RSMApp.default_log_level,
    log_format: str = "rsm",
    log_time: bool = True,
    log_lineno: bool = True,
    asset_resolver=None,
    structured: bool = False,
    standalone: bool = False,
    write_output: bool = False,
    output_dir: str = ".",
    output_filename: str = "index.html",
    custom_css: str | None = None,
) -> str | dict:
    """Process RSM source and optionally write output files.

    Parameters
    ----------
    source : str
        RSM source as a string
    path : str
        Path to RSM source file
    handrails : bool
        Include interactive handrails in output
    lint : bool
        Run linter before building
    loglevel : int
        Logging level
    log_format : str
        Format for log messages
    log_time : bool
        Include timestamps in logs
    log_lineno : bool
        Include line numbers in logs
    asset_resolver : AssetResolver, optional
        Custom asset resolver
    structured : bool
        Return structured dict instead of HTML string
    standalone : bool
        Generate standalone HTML with CDN URLs
    write_output : bool
        Write output files to disk (index.html + static/).
        When False (default), returns HTML without writing files.
        When True, writes files and still returns HTML.
    output_dir : str
        Directory where output files should be written (default: ".")
    output_filename : str
        Name of the main HTML file (default: "index.html")

    Returns
    -------
    str or dict
        HTML string (or structured dict if structured=True)
    """
    html_result = FullBuildApp(
        srcpath=path,
        plain=source,
        run_linter=lint,
        loglevel=loglevel,
        log_format=log_format,
        log_time=log_time,
        log_lineno=log_lineno,
        asset_resolver=asset_resolver,
        standalone=standalone,
        write_output=write_output,
        output_dir=output_dir,
        output_filename=output_filename,
        custom_css=custom_css,
    ).run()

    if not structured:
        return html_result

    # Parse HTML into structured components
    return _parse_html_to_structured(html_result)



from conftest import compare_have_want


def test_single():
    compare_have_want(
        have="""\
        # My Title

        This has a citation at the end. :cite:knuth::

        :references:
          @book{knuth,
            title={Art of computer programming, volume 2, Seminumerical algorithms},
            author={Knuth, Donald E},
            year={2014},
            publisher={Addison-Wesley Professional}
          }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This has a citation at the end. [<a id="cite-0" class="reference cite" href="#knuth">1</a>]</p>

        </div>

        <section class="level-2">

        <h2>References</h2>

        <ol class="bibliography" data-nodeid="4">

        <li id="knuth" class="bibitem" data-nodeid="5">
        1. Knuth, Donald E. "Art of computer programming, volume 2, Seminumerical algorithms". Addison-Wesley Professional. 2014.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        </ol>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_with_shortcuts():
    compare_have_want(
        have="""\
        # My Title

        This has a **citation** at the end. :cite:knuth::

        :references:
        @book{knuth,
          title={Art of computer programming, volume 2: Seminumerical algorithms},
          author={Knuth, Donald E},
          year={2014},
          publisher={Addison-Wesley Professional}
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This has a <span class="span" data-nodeid="3"><strong>citation</strong></span> at the end. [<a id="cite-0" class="reference cite" href="#knuth">1</a>]</p>

        </div>

        <section class="level-2">

        <h2>References</h2>

        <ol class="bibliography" data-nodeid="7">

        <li id="knuth" class="bibitem" data-nodeid="8">
        1. Knuth, Donald E. "Art of computer programming, volume 2: Seminumerical algorithms". Addison-Wesley Professional. 2014.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        </ol>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_many():
    compare_have_want(
        have="""\
        # My Title

        This has a citation at the end. :cite:torres2020,knuth::

        :references:
        @article{torres2020,
          title={Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization},
          author={Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina},
          journal={SIAM Journal on Mathematics of Data Science},
          volume={3},
          number={2},
          year={2021},
        }

        @book{knuth,
          title={Art of computer programming, volume 2: Seminumerical algorithms},
          author={Knuth, Donald E},
          year={2014},
          publisher={Addison-Wesley Professional}
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This has a citation at the end. [<span id="cite-0"><a id="cite-0-0" class="reference cite" href="#torres2020">1</a>, <a id="cite-0-1" class="reference cite" href="#knuth">2</a></span>]</p>

        </div>

        <section class="level-2">

        <h2>References</h2>

        <ol class="bibliography" data-nodeid="4">

        <li id="torres2020" class="bibitem" data-nodeid="5">
        1. Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina. "Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization". SIAM Journal on Mathematics of Data Science. 2021.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        <li id="knuth" class="bibitem" data-nodeid="6">
        2. Knuth, Donald E. "Art of computer programming, volume 2: Seminumerical algorithms". Addison-Wesley Professional. 2014.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        </ol>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_order():
    compare_have_want(
        have="""\
        # My Title

        This has a citation at the end. :cite:knuth,torres2020::

        :references:
        @article{torres2020,
          title={Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization},
          author={Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina},
          journal={SIAM Journal on Mathematics of Data Science},
          volume={3},
          number={2},
          year={2021},
        }

        @book{knuth,
          title={Art of computer programming, volume 2: Seminumerical algorithms},
          author={Knuth, Donald E},
          year={2014},
          publisher={Addison-Wesley Professional}
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This has a citation at the end. [<span id="cite-0"><a id="cite-0-0" class="reference cite" href="#knuth">2</a>, <a id="cite-0-1" class="reference cite" href="#torres2020">1</a></span>]</p>

        </div>

        <section class="level-2">

        <h2>References</h2>

        <ol class="bibliography" data-nodeid="4">

        <li id="torres2020" class="bibitem" data-nodeid="5">
        1. Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina. "Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization". SIAM Journal on Mathematics of Data Science. 2021.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        <li id="knuth" class="bibitem" data-nodeid="6">
        2. Knuth, Donald E. "Art of computer programming, volume 2: Seminumerical algorithms". Addison-Wesley Professional. 2014.<br />[<a class="reference backlink" href="#cite-0">↖1</a>]
        </li>

        </ol>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



import rsm


def test_translator_starts_anew_each_time():
    text = rsm.nodes.Text("foo")
    tr = rsm.translator.Translator()
    assert tr.translate(text) == "foo"
    assert tr.translate(text) == "foo"



"""Input: WebManuscript -- Output: None (writes to disk)."""

import logging
from pathlib import Path

from fs.copy import copy_fs

from .manuscript import WebManuscript

logger = logging.getLogger("RSM").getChild("write")


class Writer:
    """Take a WebManuscript and write to disk."""

    def __init__(self, dstpath: Path | None = None) -> None:
        self.web: WebManuscript | None = None
        self.dstpath = Path() if dstpath is None else dstpath

    def write(self, web: WebManuscript) -> None:
        self.web = web
        # Create output directory if it doesn't exist
        if not self.dstpath.exists():
            logger.info(f"Creating output directory: {self.dstpath}")
            self.dstpath.mkdir(parents=True, exist_ok=True)
        copy_fs(web, str(self.dstpath))



from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""\
        # My Title {
          :label: mylbl
          :date: 2022-03-29
        }

        :author: {
          :name: Leo Torres
          :affiliation: Max Planck Institute for Mathematics in the Sciences
          :email: leo@leotrs.com
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div id="mylbl" class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <p class="manuscript-date">March 29, 2022</p>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Leo Torres</p>

        <p>Max Planck Institute for Mathematics in the Sciences</p>

        <p><a href="mailto:leo@leotrs.com">leo@leotrs.com</a></p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_empty_author():
    compare_have_want(
        have="""\
        # The Perron non-backtracking eigenvalue after node addition {
          :label: mylbl
          :date: 2022-03-29
        }

        :author: ::

        Lorem ipsum.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div id="mylbl" class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>The Perron non-backtracking eigenvalue after node addition</h1>

        <p class="manuscript-date">March 29, 2022</p>

        <div class="author" data-nodeid="1">

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>Lorem ipsum.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_author_with_orcid():
    compare_have_want(
        have="""\
        # My Title

        :author: {
          :name: Leo Torres
          :affiliation: Some University
          :orcid: 0000-0001-2345-6789
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Leo Torres</p>

        <p>Some University</p>

        <p>0000-0001-2345-6789</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_author_with_note():
    compare_have_want(
        have="""\
        # My Title

        :author: {
          :name: Leo Torres
          :author-note: Equal contribution
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Leo Torres</p>

        <p>Equal contribution</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_author_multiline_affiliation():
    compare_have_want(
        have="""\
        # My Title

        :author: {
          :name: Leo Torres
          :affiliation: Department of Mathematics
            University of Somewhere
            Building 123
        }
        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="author" data-nodeid="1">

        <div class="paragraph">

        <p>Leo Torres</p>

        <p>Department of Mathematics</p>
        </div>
        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""
linter.py
---------

RSM Linter: analyze the manuscript tree and log linting messages.

"""

import logging

from . import nodes

logger = logging.getLogger("RSM")


class Linter:
    LINT_LVL = 25

    def __init__(self) -> None:
        logging.LINT = self.LINT_LVL  # type: ignore
        logging.addLevelName(self.LINT_LVL, "LINT")
        logger.lint = lambda msg, *args, **kwargs: logger.log(  # type: ignore
            self.LINT_LVL, msg, *args, **kwargs
        )
        if logger.level > self.LINT_LVL:
            logger.level = self.LINT_LVL

    def lint(self, tree: nodes.Manuscript) -> nodes.Manuscript:
        """Lint the manuscript."""
        logger.info("Linting...")
        self.tree: nodes.Manuscript = tree

        self.lint_manuscript_title()
        self.lint_section_label()
        self.lint_mathblock_inside_paragraph()

        return self.tree

    def _extra(self, node: nodes.Node) -> dict:
        """Make the extra arguments for the logger.lint call."""
        # Most editors start counting lines of code at 1, but TS starts at 0
        return {
            "start_row": node.start_point[0] + 1,
            "start_col": node.start_point[1] + 1,
        }

    def lint_manuscript_title(self) -> None:
        """Manuscript should have a title."""
        if not self.tree.title:
            logger.lint("Manuscript with no title", extra=self._extra(self.tree))

    def lint_section_label(self) -> None:
        """Sections should be labeled."""
        for section in self.tree.traverse(nodeclass=nodes.Section):
            if not section.label:
                logger.lint("Section with no label", extra=self._extra(section))

    def lint_mathblock_inside_paragraph(self) -> None:
        for block in self.tree.traverse(nodeclass=nodes.MathBlock):
            if not isinstance(block.parent, nodes.Paragraph):
                cls = type(block.parent)
                logger.lint(
                    f"MathBlock must be direct child of Paragraph, not {cls}",
                    extra=self._extra(block),
                )



from conftest import compare_have_want


def test_real_life_example():
    compare_have_want(
        have=r"""
        :algorithm:
        \begin{algorithm}
        \caption{Quicksort}
        \begin{algorithmic}
            \PROCEDURE{Quicksort}{$A, p, r$}
                \IF{$p < r$}
                    \STATE $q = $ \CALL{Partition}{$A, p, r$}
                    \STATE \CALL{Quicksort}{$A, p, q - 1$}
                    \STATE \CALL{Quicksort}{$A, q + 1, r$}
                \ENDIF
            \ENDPROCEDURE
            \PROCEDURE{Partition}{$A, p, r$}
                \STATE $x = A[r]$
                \STATE $i = p - 1$
                \FOR{$j = p$ \TO $r - 1$}
                    \IF{$A[j] < x$}
                        \STATE $i = i + 1$
                        \STATE exchange
                        $A[i]$ with $A[j]$
                    \ENDIF
                    \STATE exchange $A[i]$ with $A[r]$
                \ENDFOR
            \ENDPROCEDURE
        \end{algorithmic}
        \end{algorithm}
        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="algorithm">

        <pre class="pseudocode">
        \begin{algorithm}
        \caption{Quicksort}
        \begin{algorithmic}
            \PROCEDURE{Quicksort}{$A, p, r$}
                \IF{$p < r$}
                    \STATE $q = $ \CALL{Partition}{$A, p, r$}
                    \STATE \CALL{Quicksort}{$A, p, q - 1$}
                    \STATE \CALL{Quicksort}{$A, q + 1, r$}
                \ENDIF
            \ENDPROCEDURE
            \PROCEDURE{Partition}{$A, p, r$}
                \STATE $x = A[r]$
                \STATE $i = p - 1$
                \FOR{$j = p$ \TO $r - 1$}
                    \IF{$A[j] < x$}
                        \STATE $i = i + 1$
                        \STATE exchange
                        $A[i]$ with $A[j]$
                    \ENDIF
                    \STATE exchange $A[i]$ with $A[r]$
                \ENDFOR
            \ENDPROCEDURE
        \end{algorithmic}
        \end{algorithm}
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_label():
    compare_have_want(
        have=r"""
        :algorithm: {:label: alg}

        \begin{algorithm}
        \caption{Quicksort}
        \begin{algorithmic}
        k=v
        \end{algorithmic}
        \end{algorithm}
        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div id="alg" class="algorithm">

        <pre class="pseudocode">
        \begin{algorithm}
        \caption{Quicksort}
        \begin{algorithmic}
        k=v
        \end{algorithmic}
        \end{algorithm}
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""Asset resolution protocol and implementations for RSM.

This module provides a protocol-based dependency injection system for asset loading,
allowing external callers to provide custom asset loading logic while keeping the
current disk-based behavior as the default.

Technical Debt Note:
This implementation creates an architectural inconsistency where the main RSM source
file is read from disk via the Reader class, but assets are read via the AssetResolver
system. This should be unified in future refactoring to have consistent file access
patterns across the entire RSM package.
"""

import logging
from typing import Protocol

logger = logging.getLogger("RSM").getChild("asset")


class AssetResolver(Protocol):
    """Protocol for resolving asset content from various sources.

    This protocol allows RSM to obtain asset content without being coupled to
    any specific storage mechanism (filesystem, database, remote storage, etc.).
    """

    def resolve_asset(self, path: str) -> str | None:
        """Resolve an asset path to its content.

        Parameters
        ----------
        path
            The asset path/identifier to resolve.

        Returns
        -------
        Optional[str]
            The asset content as a string, or None if the asset cannot be found
            or read.
        """
        ...


class AssetResolverFromDisk:
    """Default asset resolver that reads assets from the filesystem.

    This implementation preserves the existing RSM behavior of reading assets
    directly from disk.
    """

    def resolve_asset(self, path: str) -> str | None:
        """Read asset content from filesystem.

        Parameters
        ----------
        path
            Filesystem path to the asset file.

        Returns
        -------
        Optional[str]
            File content as UTF-8 string, or None if file not found or cannot
            be read as text.
        """
        try:
            with open(path, encoding="utf-8") as f:
                return f.read()
        except FileNotFoundError:
            logger.error(f"Asset file not found: {path}")
            return None
        except (OSError, UnicodeDecodeError) as e:
            logger.error(f"Error reading asset file {path}: {e}")
            return None



from conftest import compare_have_want


def test_numbered_sections():
    compare_have_want(
        have="""\
        ## First

        Content of first.

        ## Second

        Content of second.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. First</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Content of first.</p>

        </div>

        </section>

        <section class="section level-2" data-nodeid="4">

        <h2>2. Second</h2>

        <div class="paragraph" data-nodeid="5">

        <p>Content of second.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_nonum():
    compare_have_want(
        have="""\
        ## First

        Content of first.

        ## Second {
          :nonum:
        }

        Content of second.

        ## Third

        Content of third.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. First</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Content of first.</p>

        </div>

        </section>

        <section class="section level-2" data-nodeid="4">

        <h2>Second</h2>

        <div class="paragraph" data-nodeid="5">

        <p>Content of second.</p>

        </div>

        </section>

        <section class="section level-2" data-nodeid="7">

        <h2>2. Third</h2>

        <div class="paragraph" data-nodeid="8">

        <p>Content of third.</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_one_span():
    compare_have_want(
        have="""\
        This is a :span: {:strong:} boring :: paragraph.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a <span class="span" data-nodeid="3"><strong>boring</strong></span> paragraph.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_spans():
    compare_have_want(
        have="""\
        This is a :span: {:strong:} boring :: paragraph, though it has two :span:
        {:emphas:} spans :: so it's less boring.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a <span class="span" data-nodeid="3"><strong>boring</strong></span> paragraph, though it has two <span class="span" data-nodeid="6"><em>spans</em></span> so it's less boring.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_nested():
    compare_have_want(
        have="""\
        This is a :span: {:strong:} paragraph with a :span: {:emphas:} span within a span ::
        :: so that makes it really cool.  BTW it also has a Halmos at the start of a
        line!
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a <span class="span" data-nodeid="3"><strong>paragraph with a <span class="span" data-nodeid="5"><em>span within a span</em></span></strong></span> so that makes it really cool.  BTW it also has a Halmos at the start of a line!</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_with_multiline_meta():
    compare_have_want(
        have="""\
        This is a paragraph
        with :span: {:strong:,
        :label: lbl} a
        span :: and it is multi line.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a paragraph with <span id="lbl" class="span" data-nodeid="3"><strong>a span</strong></span> and it is multi line.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_with_label():
    compare_have_want(
        have="""\
        This is a :span: {:label: myspn, :strong:} boring :: paragraph.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is a <span id="myspn" class="span" data-nodeid="3"><strong>boring</strong></span> paragraph.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_part_of_word():
    compare_have_want(
        have="""\
        This word is half bold :span: {:strong:} bo::ring.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This word is half bold <span class="span" data-nodeid="3"><strong>bo</strong></span>ring.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_span_part_of_word_with_shortcut():
    compare_have_want(
        have="""\
        This word is half bold **bo**ring.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This word is half bold <span class="span" data-nodeid="3"><strong>bo</strong></span>ring.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



import pytest

import rsm
from conftest import compare_have_want


def test_duplicate_label_warning(caplog):
    compare_have_want(
        have="""\
        # Title

        There are :span: {:label: mylbl} two :: spans with the :span: {:label: mylbl}
        same :: label in this paragraph.
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>There are <span id="mylbl" class="span" data-nodeid="3">two</span> spans with the <span class="span" data-nodeid="6">same</span> label in this paragraph.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )
    assert "Duplicate label mylbl" in caplog.text


def test_duplicate_bibtex_item_warning(caplog):
    compare_have_want(
        have="""\
        :references:
        @article{torres2020,
          title={Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization},
          author={Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina},
          journal={SIAM Journal on Mathematics of Data Science},
          year={2021},
        }

        @book{torres2020,
          title={Foo},
          author={Bar},
          journal={Baz},
          year={Bug},
        }
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="level-2">

        <h2>References</h2>

        <ol class="bibliography" data-nodeid="1">

        <li id="torres2020" class="bibitem" data-nodeid="2">
        1. Torres, Leo and Chan, Kevin S and Tong, Hanghang and Eliassi-Rad, Tina. "Nonbacktracking eigenvalues under node removal: X-centrality and targeted immunization". SIAM Journal on Mathematics of Data Science. 2021.
        </li>

        <li class="bibitem" data-nodeid="3">
        2. Bar. "Foo". Bug.
        </li>

        </ol>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )
    assert "Duplicate label torres2020" in caplog.text


def test_theorem_within_section():
    compare_have_want(
        have="""\
        ## Section

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <div class="theorem" data-nodeid="2">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_theorems_same_section():
    compare_have_want(
        have="""\
        ## Section

        :theorem:

        ::

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <div class="theorem" data-nodeid="2">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        <div class="theorem" data-nodeid="3">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.2.</span></p>

        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_theorems_different_sections():
    compare_have_want(
        have="""\
        ## Section 1

        :theorem:

        ::

        ## Section 2

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section 1</h2>

        <div class="theorem" data-nodeid="2">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        </section>

        <section class="section level-2" data-nodeid="3">

        <h2>2. Section 2</h2>

        <div class="theorem" data-nodeid="4">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 2.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_theorems_same_section_nonum():
    compare_have_want(
        have="""\
        ## Section

        :theorem:
        {:nonum:}

        ::

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <div class="theorem" data-nodeid="2">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem.</span></p>

        </div>

        </div>

        <div class="theorem" data-nodeid="3">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_theorem_inside_section_with_nonum():
    compare_have_want(
        have="""\
        ## Section
        {:nonum:}

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>Section</h2>

        <div class="theorem" data-nodeid="2">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_theorem_inside_subsection():
    compare_have_want(
        have="""\
        ## Section

        ### Subsection

        :theorem:

        ::
        """,
        want="""\
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Section</h2>

        <section class="subsection level-3" data-nodeid="2">

        <h3>1.1. Subsection</h3>

        <div class="theorem" data-nodeid="3">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_transformer_init_default():
    """Transformer can be initialized without root_dir or src_file."""
    from rsm.transformer import Transformer

    t = Transformer()
    assert t.root_dir is None
    assert t.src_file is None
    assert t.external_manuscripts == {}


def test_transformer_init_with_root_dir():
    """Transformer can be initialized with root_dir."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path("/some/root/directory")
    t = Transformer(root_dir=root_dir)
    assert t.root_dir == root_dir
    assert t.src_file is None
    assert t.external_manuscripts == {}


def test_transformer_init_with_src_file():
    """Transformer can be initialized with src_file."""
    from pathlib import Path

    from rsm.transformer import Transformer

    src_file = Path("/some/file.rsm")
    t = Transformer(src_file=src_file)
    assert t.root_dir is None
    assert t.src_file == src_file
    assert t.external_manuscripts == {}


def test_transformer_init_with_both():
    """Transformer can be initialized with both root_dir and src_file."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path("/root/dir")
    src_file = Path("/root/dir/my_file.rsm")
    t = Transformer(root_dir=root_dir, src_file=src_file)
    assert t.root_dir == root_dir
    assert t.src_file == src_file
    assert t.external_manuscripts == {}


def test_load_external_manuscript_no_root_dir():
    """_load_external_manuscript should raise error if root_dir not set."""
    from rsm.transformer import Transformer

    t = Transformer()
    with pytest.raises(ValueError, match="root_dir must be set"):
        t._load_external_manuscript("definitions/def.rsm")


def test_load_external_manuscript_file_not_found():
    """_load_external_manuscript should raise FileNotFoundError if file doesn't exist."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    with pytest.raises(FileNotFoundError, match="nonexistent.rsm"):
        t._load_external_manuscript("definitions/nonexistent.rsm")


def test_load_external_manuscript_success():
    """_load_external_manuscript should successfully load and parse external file."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    manuscript, labels_map = t._load_external_manuscript("definitions/def.rsm")

    assert manuscript is not None
    assert "test-def" in labels_map
    assert labels_map["test-def"].label == "test-def"


def test_load_external_manuscript_caching():
    """_load_external_manuscript should cache results."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    # Load first time
    manuscript1, labels1 = t._load_external_manuscript("definitions/def.rsm")

    # Verify it's cached
    assert "definitions/def.rsm" in t.external_manuscripts

    # Load second time
    manuscript2, labels2 = t._load_external_manuscript("definitions/def.rsm")

    # Should be the same objects (from cache)
    assert manuscript1 is manuscript2
    assert labels1 is labels2


def test_label_to_node_internal():
    """_label_to_node should work for internal labels (existing behavior)."""
    from rsm.transformer import Transformer

    src = """
## Section {:label: my-label}

Content here.
"""

    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)

    t = Transformer()
    t.transform(tree)

    # Internal label lookup
    node = t._label_to_node("my-label")
    assert node.label == "my-label"
    assert node.__class__.__name__ == "Section"


def test_label_to_node_external():
    """_label_to_node should load external file and find label."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    # Need to initialize the transformer by transforming an empty tree
    src = ""
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)
    t.transform(tree)

    # External label lookup
    node = t._label_to_node("test-def", external_file="definitions/def.rsm")
    assert node.label == "test-def"
    assert node.__class__.__name__ == "Section"


def test_label_to_node_external_not_found():
    """_label_to_node should return Error node if external label not found."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    # Need to initialize the transformer
    src = ""
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)
    t.transform(tree)

    # External label that doesn't exist
    node = t._label_to_node("nonexistent-label", external_file="definitions/def.rsm")
    assert node.__class__.__name__ == "Error"


def test_label_to_node_external_file_not_found():
    """_label_to_node should return Error node if external file not found."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"
    t = Transformer(root_dir=root_dir)

    # Need to initialize the transformer
    src = ""
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)
    t.transform(tree)

    # External file that doesn't exist
    node = t._label_to_node("any-label", external_file="nonexistent/file.rsm")
    assert node.__class__.__name__ == "Error"


def test_resolve_pending_references_external():
    """resolve_pending_references should resolve external references."""
    from pathlib import Path

    from rsm.transformer import Transformer

    root_dir = Path(__file__).parent / "fixtures/crossref"

    # Parse the main.rsm file which contains an external reference
    src = """# Main Document

This document references :ref:definitions/def.rsm#test-def::.
"""
    parser = rsm.tsparser.TSParser()
    tree = parser.parse(src)

    # Transform with root_dir set
    t = Transformer(root_dir=root_dir)
    transformed = t.transform(tree)

    # Find the reference node
    refs = [
        node
        for node in transformed.traverse()
        if node.__class__.__name__ == "Reference"
    ]
    assert len(refs) == 1

    ref = refs[0]
    assert ref.external_file == "definitions/def.rsm"
    assert ref.target.label == "test-def"
    assert ref.target.__class__.__name__ == "Section"



from conftest import compare_have_want


def test_one_strong():
    compare_have_want(
        have="""\
        # My Title

        This paragraph has a **shortcut** for strong span.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has a <span class="span" data-nodeid="3"><strong>shortcut</strong></span> for strong span.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_strong():
    compare_have_want(
        have="""\
        # My Title

        This paragraph has two **shortcuts** for strong **spans**.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has two <span class="span" data-nodeid="3"><strong>shortcuts</strong></span> for strong <span class="span" data-nodeid="6"><strong>spans</strong></span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_one_emphas():
    compare_have_want(
        have="""\
        # My Title

        This paragraph has a *shortcut* for emphasis span.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has a <span class="span" data-nodeid="3"><em>shortcut</em></span> for emphasis span.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_one_math():
    compare_have_want(
        have="""\
        # My Title

        This paragraph has some $2+2=4$ awesome math.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has some <span class="math" data-nodeid="3">\(2+2=4\)</span> awesome math.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_two_math():
    compare_have_want(
        have="""\
        # My Title

        This paragraph has some $2+2=4$ awesome math and also some math that $2 + 2 + 2
        + 2 + 2 = 10$ is broken down.
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>My Title</h1>

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has some <span class="math" data-nodeid="3">\(2+2=4\)</span> awesome math and also some math that <span class="math" data-nodeid="6">\(2 + 2 + 2
        + 2 + 2 = 10\)</span> is broken down.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_mathblock():
    compare_have_want(
        have="""\
        This paragraph is followed by display math
        $$2 + 2 = 4.$$
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph is followed by display math </p>
        <div class="mathblock" data-nodeid="3">
        $$
        2 + 2 = 4.
        $$
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_section_shortcut():
    compare_have_want(
        have="""\
        ## My Section{:label: my-sec}

        This section contains a shortcut
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section id="my-sec" class="section level-2" data-nodeid="1">

        <h2>1. My Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>This section contains a shortcut</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_subsubsection_shortcut():
    compare_have_want(
        have="""\
        ## My Section

        Foo.

        ### My Subsection

        Bar.

        #### My Subsubsection

        Baz.

        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. My Section</h2>

        <div class="paragraph" data-nodeid="2">

        <p>Foo.</p>

        </div>

        <section class="subsection level-3" data-nodeid="4">

        <h3>1.1. My Subsection</h3>

        <div class="paragraph" data-nodeid="5">

        <p>Bar.</p>

        </div>

        <section class="subsubsection level-4" data-nodeid="7">

        <h4>1.1.1. My Subsubsection</h4>

        <div class="paragraph" data-nodeid="8">

        <p>Baz.</p>

        </div>

        </section>

        </section>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_claim_shortcut():
    compare_have_want(
        have="""\
        This paragraph contains a :|-: claim with a turnstile::. And also another one :⊢: but it takes
        multiple lines::.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph contains a <span class="construct claim" data-nodeid="3"><span class="keyword" data-nodeid="4">⊢</span> claim with a turnstile</span>. And also another one <span class="construct claim" data-nodeid="8"><span class="keyword" data-nodeid="9">⊢</span> but it takes multiple lines</span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_prev_shortcut():
    compare_have_want(
        have="""\
        :theorem:

        All $X$ are $Y$.

        ::

        :proof:

        :step:{:label: st1}
          All $X$ are $Z$ and all $Z$ are $Y$.
        ::

        :step: :qed:.

           :p: Due to :prev:.::

        ::

        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="theorem" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>All <span class="math" data-nodeid="4">\(X\)</span> are <span class="math" data-nodeid="7">\(Y\)</span>.</p>

        </div>

        </div>

        <div class="proof" data-nodeid="10">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof.</span></p>

        </div>

        <div id="st1" class="step" data-nodeid="11">

        <div class="statement" data-nodeid="12">

        <div class="paragraph" data-nodeid="13">

        <p>All <span class="math" data-nodeid="15">\(X\)</span> are <span class="math" data-nodeid="18">\(Z\)</span> and all <span class="math" data-nodeid="21">\(Z\)</span> are <span class="math" data-nodeid="24">\(Y\)</span>.</p>

        </div>

        </div>

        </div>

        <div class="step last" data-nodeid="27">

        <div class="statement" data-nodeid="28">

        <div class="paragraph" data-nodeid="29">

        <p><span class="construct qed" data-nodeid="30"><span class="keyword" data-nodeid="31">QED</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="34">

        <div class="paragraph" data-nodeid="35">

        <p>Due to <a class="reference" href="#st1">Step ⟨1⟩</a>.</p>

        </div>

        </div>

        </div>

        <div class="halmos"></div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_prev2_shortcut():
    compare_have_want(
        have="""\
        :theorem:

        All $X$ are $Y$.

        ::

        :proof:

        :step:
          {:label: st1}
          All $X$ are $Z$ and all $Z$ are $Y$.
        ::

        :step:
          {:label: st2}
          Something inconsequential here.
        ::

        :step: :qed:.

           :p: Due to :prev2:.::

        ::

        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="theorem" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>All <span class="math" data-nodeid="4">\(X\)</span> are <span class="math" data-nodeid="7">\(Y\)</span>.</p>

        </div>

        </div>

        <div class="proof" data-nodeid="10">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof.</span></p>

        </div>

        <div id="st1" class="step" data-nodeid="11">

        <div class="statement" data-nodeid="12">

        <div class="paragraph" data-nodeid="13">

        <p>All <span class="math" data-nodeid="15">\(X\)</span> are <span class="math" data-nodeid="18">\(Z\)</span> and all <span class="math" data-nodeid="21">\(Z\)</span> are <span class="math" data-nodeid="24">\(Y\)</span>.</p>

        </div>

        </div>

        </div>

        <div id="st2" class="step" data-nodeid="27">

        <div class="statement" data-nodeid="28">

        <div class="paragraph" data-nodeid="29">

        <p>Something inconsequential here.</p>

        </div>

        </div>

        </div>

        <div class="step last" data-nodeid="31">

        <div class="statement" data-nodeid="32">

        <div class="paragraph" data-nodeid="33">

        <p><span class="construct qed" data-nodeid="34"><span class="keyword" data-nodeid="35">QED</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="38">

        <div class="paragraph" data-nodeid="39">

        <p>Due to <a class="reference" href="#st1">Step ⟨1⟩</a>.</p>

        </div>

        </div>

        </div>

        <div class="halmos"></div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_prev_and_prev2_shortcut():
    compare_have_want(
        have="""\
        :theorem:

        All $X$ are $Y$.

        ::

        :proof:

        :step:
          {:label: st1}
          All $X$ are $Z$.
        ::

        :step:
          {:label: st2}
          All $Z$ are $Y$.
        ::

        :step: :qed:.

           :p: Due to :prev: and :prev2:.::

        ::

        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="theorem" data-nodeid="1">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.</span></p>

        </div>

        <div class="paragraph" data-nodeid="2">

        <p>All <span class="math" data-nodeid="4">\(X\)</span> are <span class="math" data-nodeid="7">\(Y\)</span>.</p>

        </div>

        </div>

        <div class="proof" data-nodeid="10">

        <div class="paragraph hr-label">

        <p><span class="span label">Proof.</span></p>

        </div>

        <div id="st1" class="step" data-nodeid="11">

        <div class="statement" data-nodeid="12">

        <div class="paragraph" data-nodeid="13">

        <p>All <span class="math" data-nodeid="15">\(X\)</span> are <span class="math" data-nodeid="18">\(Z\)</span>.</p>

        </div>

        </div>

        </div>

        <div id="st2" class="step" data-nodeid="21">

        <div class="statement" data-nodeid="22">

        <div class="paragraph" data-nodeid="23">

        <p>All <span class="math" data-nodeid="25">\(Z\)</span> are <span class="math" data-nodeid="28">\(Y\)</span>.</p>

        </div>

        </div>

        </div>

        <div class="step last" data-nodeid="31">

        <div class="statement" data-nodeid="32">

        <div class="paragraph" data-nodeid="33">

        <p><span class="construct qed" data-nodeid="34"><span class="keyword" data-nodeid="35">QED</span></span>.</p>

        </div>

        </div>

        <div class="subproof" data-nodeid="38">

        <div class="paragraph" data-nodeid="39">

        <p>Due to <a class="reference" href="#st2">Step ⟨2⟩</a> and <a class="reference" href="#st1">Step ⟨1⟩</a>.</p>

        </div>

        </div>

        </div>

        <div class="halmos"></div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_code_shortcut():
    compare_have_want(
        have="""\
        This is inline code`comp = [abs(x) for x in range(10)]`.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This is inline code<span class="code" data-nodeid="3"><code>comp = [abs(x) for x in range(10)]</code></span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_codeblock_shortcut():
    compare_have_want(
        have="""\
        ```
        comp = [abs(x) for x in range(10)]
        ```
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>comp = [abs(x) for x in range(10)]</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_hashtag_not_at_the_start_of_line():
    compare_have_want(
        have=r"""
        ## This is a section title

        And \# this is not!
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. This is a section title</h2>

        <div class="paragraph" data-nodeid="2">

        <p>And # this is not!</p>

        </div>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_escaped_colon_behind_halmos():
    compare_have_want(
        have=r"""
        Shortcut right beside an escaped colon :span: {:emphas:} foo\:::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Shortcut right beside an escaped colon <span class="span" data-nodeid="3"><em>foo:</em></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_escaped_colon_behind_halmos_after_shortcut():
    compare_have_want(
        have=r"""
        Shortcut right beside an escaped colon **foo\:**
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Shortcut right beside an escaped colon <span class="span" data-nodeid="3"><strong>foo:</strong></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_math_after_turnstile():
    compare_have_want(
        have="""\
        This paragraph has math inside a claim :⊢: $2+2=4$::.
        """,
        want=r"""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>This paragraph has math inside a claim <span class="construct claim" data-nodeid="3"><span class="keyword" data-nodeid="4">⊢</span> <span class="math" data-nodeid="7">\(2+2=4\)</span></span>.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_turnstile_and_math_within_list():
    compare_have_want(
        have="""\
        We now make a bunch of claims

        :enumerate:

        :-: {:label: one} :⊢: $2+2=4$::.

        :-: {:label: two} :⊢: $3+3=6$::.

        ::
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>We now make a bunch of claims</p>

        </div>

        <ol class="enumerate" data-nodeid="3">

        <li id="one" class="item" data-nodeid="4">
        <span class="construct claim" data-nodeid="5"><span class="keyword" data-nodeid="6">⊢</span> <span class="math" data-nodeid="9">\(2+2=4\)</span></span>.
        </li>

        <li id="two" class="item" data-nodeid="12">
        <span class="construct claim" data-nodeid="13"><span class="keyword" data-nodeid="14">⊢</span> <span class="math" data-nodeid="17">\(3+3=6\)</span></span>.
        </li>

        </ol>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_asterisk_inside_math():
    compare_have_want(
        have="""\
        $2*2 = 4*1$
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="math" data-nodeid="2">\(2*2 = 4*1\)</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_asterisk_inside_code():
    compare_have_want(
        have="""\
        `[x**2 for x in range(10)]`
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="code" data-nodeid="2"><code>[x**2 for x in range(10)]</code></span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_asterisk_inside_mathblock():
    compare_have_want(
        have=r"""
        Here comes a mathblock with an asterisk:
        $$
        2*2 = 4*1
        $$
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Here comes a mathblock with an asterisk: </p>
        <div class="mathblock" data-nodeid="3">
        $$
        2*2 = 4*1
        $$
        </div>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_asterisk_inside_codeblock():
    compare_have_want(
        have="""\
        ```
        2*2 = 4*1
        ```
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="codeblock" data-nodeid="1">

        <pre>
        <code>2*2 = 4*1</code>
        </pre>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )



from conftest import compare_have_want


def test_simple():
    compare_have_want(
        have="""
        ## Foo

        :appendix:

        ## Bar
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Foo</h2>

        </section>

        <section class="section level-2" data-nodeid="3">

        <h2>A. Bar</h2>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_subsection():
    compare_have_want(
        have="""
        ## Foo
        ### Sub Foo

        :appendix:

        ## Bar
        ### Sub Bar
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Foo</h2>

        <section class="subsection level-3" data-nodeid="2">

        <h3>1.1. Sub Foo</h3>

        </section>

        </section>

        <section class="section level-2" data-nodeid="4">

        <h2>A. Bar</h2>

        <section class="subsection level-3" data-nodeid="5">

        <h3>A.1. Sub Bar</h3>

        </section>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_theorem():
    compare_have_want(
        have="""
        ## Foo
        ### Sub Foo

        :theorem: ::

        :appendix:

        ## Bar
        ### Sub Bar

        :theorem: ::
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <section class="section level-2" data-nodeid="1">

        <h2>1. Foo</h2>

        <section class="subsection level-3" data-nodeid="2">

        <h3>1.1. Sub Foo</h3>

        <div class="theorem" data-nodeid="3">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem 1.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        <section class="section level-2" data-nodeid="5">

        <h2>A. Bar</h2>

        <section class="subsection level-3" data-nodeid="6">

        <h3>A.1. Sub Bar</h3>

        <div class="theorem" data-nodeid="7">

        <div class="paragraph hr-label">

        <p><span class="span label">Theorem A.1.</span></p>

        </div>

        </div>

        </section>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_section_reftext():
    compare_have_want(
        have="""
        :ref:app::

        :appendix:

        ## Foo bar {:label: app}
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><a class="reference" href="#app">Appendix A</a></p>

        </div>

        <section id="app" class="section level-2" data-nodeid="4">

        <h2>A. Foo bar</h2>

        </section>

        </section>

        </div>

        </div>

        </body>
        """,
    )



"""
manuscript.py
-------------

Classes that represent the manuscript at different stages.

"""

from pathlib import Path

from fs.mountfs import MountFS


class WebManuscript(MountFS):
    def __init__(self, src: Path = None) -> None:
        super().__init__()
        self.src = Path(src) if src else None
        self.body: str = ""
        self.html: str = ""

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        classname = self.__class__.__name__
        return f"{classname}(src='{self.src}')"



from conftest import compare_have_want
from rsm.util import EscapedString


def test_ignore_single_char():
    esc = EscapedString(r"\:", ":")
    assert esc.find(":") == -1


def test_simple():
    compare_have_want(
        have=r"""
        Warning\: this is a warning.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p>Warning: this is a warning.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_colon_inside_bold():
    compare_have_want(
        have=r"""
        :span: {:strong:} Warning: :: this is a warning.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2"><strong>Warning:</strong></span> this is a warning.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_colon_inside_bold_no_space():
    compare_have_want(
        have=r"""
        :span: {:strong:} Warning: :: this is a warning.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2"><strong>Warning:</strong></span> this is a warning.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_colon_inside_bold_with_shortcut():
    compare_have_want(
        have=r"""
        **Warning: **this is a warning.
        """,
        want="""
        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="span" data-nodeid="2"><strong>Warning:</strong></span>this is a warning.</p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_backslash_brace_within_math():
    compare_have_want(
        have=r"""
        $[k]:= \{1,\ldots,k\}$
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <div class="paragraph" data-nodeid="1">

        <p><span class="math" data-nodeid="2">\([k]:= \{1,\ldots,k\}\)</span></p>

        </div>

        </section>

        </div>

        </div>

        </body>
        """,
    )


def test_colon_inside_manuscript_title():
    compare_have_want(
        have=r"""
        # foo: bar
        """,
        want=r"""        <body>

        <div class="manuscriptwrapper">

        <div class="manuscript" data-nodeid="0">

        <section class="level-1">

        <h1>foo: bar</h1>

        </section>

        </div>

        </div>

        </body>
        """,
    )
